
import os
import tempfile
from http.cookiejar import CookieJar
from pathlib import Path
from typing import Dict, List, Optional, Union
from pycsghub.utils import (get_file_download_url,
                            model_id_to_group_owner_name)
from pycsghub.cache import ModelFileSystemCache
from pycsghub.utils import (get_cache_dir,
                            pack_repo_file_info,
                            get_endpoint)
from huggingface_hub.utils import filter_repo_objects
from pycsghub.file_download import http_get
from pycsghub.constants import DEFAULT_REVISION, REPO_TYPES
from pycsghub import utils
from pycsghub.constants import REPO_TYPE_MODEL

def snapshot_download(
        repo_id: str,
        *,
        repo_type: Optional[str] = None,
        revision: Optional[str] = DEFAULT_REVISION,
        cache_dir: Union[str, Path, None] = None,
        local_files_only: Optional[bool] = False,
        cookies: Optional[CookieJar] = None,
        allow_patterns: Optional[Union[List[str], str]] = None,
        ignore_patterns: Optional[Union[List[str], str]] = None,
        headers: Optional[Dict[str, str]] = None,
        endpoint: Optional[str] = None,
        token: Optional[str] = None
) -> str:
    if repo_type is None:
        repo_type = REPO_TYPE_MODEL
    if repo_type not in REPO_TYPES:
        raise ValueError(f"Invalid repo type: {repo_type}. Accepted repo types are: {str(REPO_TYPES)}")
    if cache_dir is None:
        cache_dir = get_cache_dir(repo_type=repo_type)
    if isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)
    temporary_cache_dir = os.path.join(cache_dir, 'temp')
    os.makedirs(temporary_cache_dir, exist_ok=True)

    group_or_owner, name = model_id_to_group_owner_name(repo_id)
    # name = name.replace('.', '___')

    cache = ModelFileSystemCache(cache_dir, group_or_owner, name)

    if local_files_only:
        if len(cache.cached_files) == 0:
            raise ValueError(
                'Cannot find the requested files in the cached path and outgoing'
                ' traffic has been disabled. To enable model look-ups and downloads'
                " online, set 'local_files_only' to False.")
        return cache.get_root_location()
    else:
        download_endpoint = get_endpoint(endpoint=endpoint)
        # make headers
        # todo need to add cookies？
        repo_info = utils.get_repo_info(repo_id,
                                        repo_type=repo_type,
                                        revision=revision,
                                        token=token,
                                        endpoint=download_endpoint)

        assert repo_info.sha is not None, "Repo info returned from server must have a revision sha."
        assert repo_info.siblings is not None, "Repo info returned from server must have a siblings list."
        repo_files = list(
            filter_repo_objects(
                items=[f.rfilename for f in repo_info.siblings],
                allow_patterns=allow_patterns,
                ignore_patterns=ignore_patterns,
            )
        )

        with tempfile.TemporaryDirectory(dir=temporary_cache_dir) as temp_cache_dir:
            for repo_file in repo_files:
                repo_file_info = pack_repo_file_info(repo_file, revision)
                if cache.exists(repo_file_info):
                    file_name = os.path.basename(repo_file_info['Path'])
                    print(f"File {file_name} already in cache '{cache.get_root_location()}', skip downloading!")
                    continue

                # get download url
                url = get_file_download_url(
                    model_id=repo_id,
                    file_path=repo_file,
                    repo_type=repo_type,
                    revision=revision,
                    endpoint=download_endpoint)
                # todo support parallel download api
                http_get(
                    url=url,
                    local_dir=temp_cache_dir,
                    file_name=repo_file,
                    headers=headers,
                    cookies=cookies,
                    token=token)

                # todo using hash to check file integrity
                temp_file = os.path.join(temp_cache_dir, repo_file)
                savedFile = cache.put_file(repo_file_info, temp_file)
                print(f"Saved file to '{savedFile}'")
            
        cache.save_model_version(revision_info={'Revision': revision})
        return os.path.join(cache.get_root_location())



import typer
from enum import Enum
from pycsghub.constants import REPO_TYPE_MODEL, REPO_TYPE_DATASET, REPO_TYPE_SPACE

class RepoType(str, Enum):
    MODEL = REPO_TYPE_MODEL
    DATASET = REPO_TYPE_DATASET
    SPACE = REPO_TYPE_SPACE



import re

_asciire = re.compile('([\x00-\x7f]+)')
_hexdig = '0123456789ABCDEFabcdef'
_hextobyte = None

def unquote_to_bytes(string):
    """unquote_to_bytes('abc%20def') -> b'abc def'."""
    # Note: strings are encoded as UTF-8. This is only an issue if it contains
    # unescaped non-ASCII characters, which URIs should not.
    if not string:
        # Is it a string-like object?
        string.split
        return b''
    if isinstance(string, str):
        string = string.encode('utf-8')
    bits = string.split(b'%')
    if len(bits) == 1:
        return string
    res = [bits[0]]
    append = res.append
    # Delay the initialization of the table to not waste memory
    # if the function is never called
    global _hextobyte
    if _hextobyte is None:
        _hextobyte = {(a + b).encode(): bytes.fromhex(a + b)
                      for a in _hexdig for b in _hexdig}
    for item in bits[1:]:
        try:
            append(_hextobyte[item[:2]])
            append(item[2:])
        except KeyError:
            append(b'%')
            append(item)
    return b''.join(res)

def unquote(string, encoding='utf-8', errors='replace'):
    """Replace %xx escapes by their single-character equivalent. The optional
    encoding and errors parameters specify how to decode percent-encoded
    sequences into Unicode characters, as accepted by the bytes.decode()
    method.
    By default, percent-encoded sequences are decoded with UTF-8, and invalid
    sequences are replaced by a placeholder character.

    unquote('abc%20def') -> 'abc def'.
    """
    if isinstance(string, bytes):
        return unquote_to_bytes(string).decode(encoding, errors)
    if '%' not in string:
        string.split
        return string
    if encoding is None:
        encoding = 'utf-8'
    if errors is None:
        errors = 'replace'
    bits = _asciire.split(string)
    res = [bits[0]]
    append = res.append
    for i in range(1, len(bits), 2):
        append(unquote_to_bytes(bits[i]).decode(encoding, errors))
        append(bits[i + 1])
    return ''.join(res)



import logging
import os
import time
from configparser import ConfigParser
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Dict
from .fixes import WeakFileLock

logger = logging.getLogger(__name__)

key_timestamp = "timestamp"
key_size = "size"
key_should_ignore = "should_ignore"
key_sha256 = "sha256"
key_sha1 = "sha1"
key_upload_mode = "upload_mode"
key_is_uploaded = "is_uploaded"
key_is_committed = "is_committed"
key_lfs_upload_id = "lfs_upload_id"
key_lfs_part_count = "lfs_part_count"
key_lfs_uploaded_ids = "lfs_uploaded_ids"
key_remote_oid = "remote_oid"

@dataclass(frozen=True)
class LocalUploadFilePaths:
    path_in_repo: str
    file_path: Path
    lock_path: Path
    metadata_path: Path
    
@dataclass
class LocalUploadFileMetadata:
    """Metadata for a file that is being uploaded to the hub."""
    size: int

    timestamp: Optional[float] = None
    should_ignore: Optional[bool] = None
    sha256: Optional[str] = None
    sha1: Optional[str] = None
    upload_mode: Optional[str] = None  # regular | lfs
    is_uploaded: bool = False
    is_committed: bool = False
    
    remote_oid: Optional[str] = None # remote oid
    lfs_upload_id: Optional[str] = None # upload id, only used for multipart uploads
    lfs_uploaded_ids: Optional[str] = None # uploaded ids

    # only for runtime
    lfs_upload_part_count: Optional[int] = None # total number of parts, only used for multipart uploads
    lfs_upload_part_index: Optional[int] = None # index of part, only used for multipart uploads
    lfs_upload_part_url: Optional[str] = None # upload url for part
    lfs_upload_chunk_size: Optional[int] = None
    lfs_upload_complete_url: Optional[str] = None # for merge multi-part
    lfs_upload_verify: Optional[Dict] = None # for verify
    content_base64: str = ""
    
    def save(self, paths: LocalUploadFilePaths) -> None:
        """Save the metadata to disk."""
        with WeakFileLock(paths.lock_path):
            new_timestamp = time.time()
            metadata = {
                key_timestamp: time.time(),
                key_size: str(self.size),
                key_should_ignore: "" if self.should_ignore is None else str(self.should_ignore),
                key_sha256: "" if self.sha256 is None else self.sha256,
                key_sha1: "" if self.sha1 is None else self.sha1,
                key_upload_mode: "" if self.upload_mode is None else self.upload_mode,
                key_is_uploaded: str(self.is_uploaded),
                key_is_committed: str(self.is_committed),
                key_remote_oid: "" if self.remote_oid is None else self.remote_oid,
                key_lfs_upload_id: "" if self.lfs_upload_id is None else self.lfs_upload_id,
                key_lfs_part_count: "" if self.lfs_upload_part_count is None else str(self.lfs_upload_part_count),
                key_lfs_uploaded_ids: "" if self.lfs_uploaded_ids is None else self.lfs_uploaded_ids,
            }
            save_properties(paths.metadata_path, metadata)
            self.timestamp = new_timestamp

def read_upload_metadata(local_dir: Path, filename: str) -> LocalUploadFileMetadata:
    paths = get_local_upload_paths(local_dir, filename)
    with WeakFileLock(paths.lock_path):
        if paths.metadata_path.exists():
            try:
                props = read_properties(paths.metadata_path)
                
                timestamp = float(props.get(key_timestamp))
                
                size = int(props.get(key_size))
                
                _should_ignore = props.get(key_should_ignore)
                should_ignore = None if _should_ignore == "" else _should_ignore.lower() == "true"
                
                sha256 = props.get(key_sha256)
                sha1 = props.get(key_sha1)
                
                _upload_mode = props.get(key_upload_mode)
                upload_mode = None if _upload_mode == "" else _upload_mode
                
                is_uploaded = props.get(key_is_uploaded).lower() == "true"
                is_committed = props.get(key_is_committed).lower() == "true"

                _lfs_upload_id = props.get(key_lfs_upload_id)
                lfs_upload_id = None if _lfs_upload_id == "" else _lfs_upload_id
                _lfs_uploaded_ids = props.get(key_lfs_uploaded_ids)
                lfs_uploaded_ids = None if _lfs_uploaded_ids == "" else _lfs_uploaded_ids
                
                metadata = LocalUploadFileMetadata(
                    timestamp=timestamp,
                    size=size,
                    should_ignore=should_ignore,
                    sha256=sha256,
                    sha1=sha1,
                    upload_mode=upload_mode,
                    is_uploaded=is_uploaded,
                    is_committed=is_committed,
                    lfs_upload_id=lfs_upload_id,
                    lfs_uploaded_ids=lfs_uploaded_ids,
                )
            except Exception as e:
                # remove the metadata file if it is corrupted / not the right format
                logger.warning(f"invalid metadata file {paths.metadata_path}: {e}. Removing it from disk and continue.")
                try:
                    paths.metadata_path.unlink()
                except Exception as e:
                    logger.warning(f"could not remove corrupted metadata file {paths.metadata_path}: {e}")

            if (
                metadata.timestamp is not None
                and metadata.is_uploaded  # file was uploaded
                and not metadata.is_committed  # but not committed
                and time.time() - metadata.timestamp > 20 * 3600  # and it's been more than 20 hours
            ):
                metadata.is_uploaded = False

            # check if the file exists and hasn't been modified since the metadata was saved
            try:
                if metadata.timestamp is not None and paths.file_path.stat().st_mtime <= metadata.timestamp:
                    return metadata
                logger.info(f"ignored metadata for '{filename}' (outdated) and will re-compute hash.")
            except FileNotFoundError:
                # file does not exist => metadata is outdated
                pass

    # empty metadata => we don't know anything expect its size
    return LocalUploadFileMetadata(size=paths.file_path.stat().st_size)

def read_properties(file_path):
    config = ConfigParser()
    with open(file_path, 'r', encoding='utf-8') as f:
        config.read_string(f.read())
    return dict(config['DEFAULT'])

def save_properties(file_path: str, data: dict) -> None:
    config = ConfigParser()
    config['DEFAULT'] = data
    with open(file_path, 'w', encoding='utf-8') as f:
        config.write(f, space_around_delimiters=False)

def get_local_upload_paths(local_dir: Path, filename: str) -> LocalUploadFilePaths:
    sanitized_filename = os.path.join(*filename.split("/"))
    if os.name == "nt":
        if sanitized_filename.startswith("..\\") or "\\..\\" in sanitized_filename:
            raise ValueError(
                f"Invalid filename: cannot handle filename '{sanitized_filename}' on Windows. Please ask the repository"
                " owner to rename this file."
            )
    file_path = local_dir / sanitized_filename
    metadata_path = csghub_dir(local_dir) / "upload" / f"{sanitized_filename}.metadata"
    lock_path = metadata_path.with_suffix(".lock")

    # Some Windows versions do not allow for paths longer than 255 characters.
    # In this case, we must specify it as an extended path by using the "\\?\" prefix
    if os.name == "nt":
        if not str(local_dir).startswith("\\\\?\\") and len(os.path.abspath(lock_path)) > 255:
            file_path = Path("\\\\?\\" + os.path.abspath(file_path))
            lock_path = Path("\\\\?\\" + os.path.abspath(lock_path))
            metadata_path = Path("\\\\?\\" + os.path.abspath(metadata_path))

    file_path.parent.mkdir(parents=True, exist_ok=True)
    metadata_path.parent.mkdir(parents=True, exist_ok=True)
    return LocalUploadFilePaths(
        path_in_repo=filename, file_path=file_path, lock_path=lock_path, metadata_path=metadata_path
    )


def csghub_dir(local_dir: Path) -> Path:
    path = local_dir / ".cache" / "csghub"
    path.mkdir(exist_ok=True, parents=True)

    gitignore = path / ".gitignore"
    gitignore_lock = path / ".gitignore.lock"
    if not gitignore.exists():
        try:
            with WeakFileLock(gitignore_lock):
                gitignore.write_text("*")
            gitignore_lock.unlink()
        except OSError:
            pass
    return path



try:
    from requests import JSONDecodeError  # type: ignore  # noqa: F401
except ImportError:
    try:
        from simplejson import JSONDecodeError  # type: ignore # noqa: F401
    except ImportError:
        from json import JSONDecodeError  # type: ignore  # noqa: F401
import contextlib
import os
import shutil
import stat
import tempfile
from functools import partial
from pathlib import Path
from typing import Callable, Generator, Optional, Union

import yaml
from filelock import BaseFileLock, FileLock, SoftFileLock, Timeout

from . import consts
import logging

logger = logging.getLogger(__name__)

yaml_dump: Callable[..., str] = partial(yaml.dump, stream=None, allow_unicode=True)  # type: ignore

@contextlib.contextmanager
def SoftTemporaryDirectory(
    suffix: Optional[str] = None,
    prefix: Optional[str] = None,
    dir: Optional[Union[Path, str]] = None,
    **kwargs,
) -> Generator[Path, None, None]:
    tmpdir = tempfile.TemporaryDirectory(prefix=prefix, suffix=suffix, dir=dir, **kwargs)
    yield Path(tmpdir.name).resolve()

    try:
        # First once with normal cleanup
        shutil.rmtree(tmpdir.name)
    except Exception:
        # If failed, try to set write permission and retry
        try:
            shutil.rmtree(tmpdir.name, onerror=_set_write_permission_and_retry)
        except Exception:
            pass

    # And finally, cleanup the tmpdir.
    # If it fails again, give up but do not throw error
    try:
        tmpdir.cleanup()
    except Exception:
        pass


def _set_write_permission_and_retry(func, path, excinfo):
    os.chmod(path, stat.S_IWRITE)
    func(path)


@contextlib.contextmanager
def WeakFileLock(lock_file: Union[str, Path]) -> Generator[BaseFileLock, None, None]:
    lock = FileLock(lock_file, timeout=consts.FILELOCK_LOG_EVERY_SECONDS)
    while True:
        try:
            lock.acquire()
        except Timeout:
            logger.info("still waiting to acquire lock on %s", lock_file)
        except NotImplementedError as e:
            if "use SoftFileLock instead" in str(e):
                logger.warning(
                    "FileSystem does not appear to support flock. Falling back to SoftFileLock for %s", lock_file
                )
                lock = SoftFileLock(lock_file, timeout=consts.FILELOCK_LOG_EVERY_SECONDS)
                continue
        else:
            break

    yield lock

    try:
        return lock.release()
    except OSError:
        try:
            Path(lock_file).unlink()
        except OSError:
            pass
